@@ -5,7 +5,7 @@
 # Copyright (C) 2002       Masato Taruishi <taru@debian.org>
 # Copyright (C) 2006-2008  Junichi Uekawa <dancer@debian.org>
 # Copyright (C) 2007       Famelis George <famelis@otenet.gr>
-# Copyright (C) 2008-2015  Francesco Poli <invernomuto@paranoici.org>
+# Copyright (C) 2008-2016  Francesco Poli <invernomuto@paranoici.org>
 # Copyright (C) 2009-2010  Ryan Niebur <ryan@debian.org>
 # Copyright (C) 2013       Google Inc
 #
@@ -27,12 +27,12 @@
 require 'getoptlong'
 require 'debian'
 require 'unicode'
-require 'aptlistbugs/debian/bug'
 require 'aptlistbugs/debian/bts'
 require 'thread'
 require 'tempfile'
 require 'gettext'
 require 'rss/maker'
+require 'fcntl'
 require 'aptlistbugs/debian/apt_preferences'
 
 include GetText
@@ -49,8 +49,6 @@ class AppConfig
       "\n",
       _("Options:\n"),
       # TRANSLATORS: the colons (:) in the following strings are vertically aligned, please keep their alignment consistent
-      _(" -h               : Display this help and exit.\n"),
-      _(" -v               : Show version number and exit.\n"),
       # TRANSLATORS: the \"all\" between quotes should not be translated
       sprintf(_(" -s <severities>  : Filter bugs by severities you want to see\n                    (or \"all\" for all)\n                    [%s].\n"), @severity.join(',')),
       _(" -T <tags>        : Filter bugs by tags you want to see.\n"),
@@ -64,9 +62,12 @@ class AppConfig
       _(" -q               : Don't display progress bar.\n"),
       _(" -C <apt.conf>    : apt.conf file to use.\n"),
       _(" -F               : Automatically pin all buggy packages.\n"),
+      _(" -N               : Never automatically pin packages.\n"),
       _(" -y               : Assume that you select yes for all questions.\n"),
       _(" -n               : Assume that you select no for all questions.\n"),
       _(" -d               : Debug.\n"),
+      _(" -h               : Display this help and exit.\n"),
+      _(" -v               : Show version number and exit.\n"),
       _("Commands:\n"),
       _(" apt              : Apt mode.\n"),
       _(" list <pkg...>    : List bug reports of the specified packages.\n"),
@@ -89,6 +90,7 @@ class AppConfig
     @show_downgrade = false
     @hostname = "bugs.debian.org"
     @port = 80
+    @querystep = 200
     @parsestep = 200
     @quiet = false
     @command = nil
@@ -105,7 +107,7 @@ class AppConfig
     @apt_conf = nil
 
     @yes = nil
-    @force_pin = false
+    @force_pin = nil
 
     @arrow = "->"
     if Locale.charset == "UTF-8"
@@ -114,7 +116,7 @@ class AppConfig
   end
 
   attr_accessor :severity, :stats, :quiet, :title
-  attr_accessor :show_downgrade, :hostname, :tag, :fbugs, :parsestep
+  attr_accessor :show_downgrade, :hostname, :tag, :fbugs, :querystep, :parsestep
   attr_accessor :frontend, :pin_priority, :yes, :ignore_regexp, :force_pin
   attr_reader :command, :parser, :querybts, :ignore_bugs, :system_ignore_bugs, :browser, :arrow
 
@@ -128,16 +130,12 @@ class AppConfig
       end
     end
 
-    if /qb='(.*)'/ =~ `apt-config #{@apt_conf} shell qb AptListbugs::ParseStep`
-      @parsestep = $1.to_i if $1.to_i > 0
+    if /qb='(.*)'/ =~ `apt-config #{@apt_conf} shell qb AptListbugs::QueryStep`
+      @querystep = $1.to_i if $1.to_i > 0
     end
 
-    # this hack is not nice: it should be dropped, once
-    # Debian support for Ruby 1.8 is long gone
-    getoptlong_ambiguousoption = if RUBY_VERSION >= "1.9"
-      GetoptLong::AmbiguousOption
-    else
-      GetoptLong::AmbigousOption
+    if /qb='(.*)'/ =~ `apt-config #{@apt_conf} shell qb AptListbugs::ParseStep`
+      @parsestep = $1.to_i if $1.to_i > 0
     end
 
     opt_parser = GetoptLong.new
@@ -155,6 +153,7 @@ class AppConfig
                            ['--quiet', '-q', GetoptLong::NO_ARGUMENT],
                            ['--aptconf', '-C', GetoptLong::REQUIRED_ARGUMENT],
                            ['--force-pin', '-F', GetoptLong::NO_ARGUMENT],
+                           ['--force-no-pin', '-N', GetoptLong::NO_ARGUMENT],
                            ['--force-yes', '-y', GetoptLong::NO_ARGUMENT],
                            ['--force-no', '-n', GetoptLong::NO_ARGUMENT],
                            ['--debug', '-d', GetoptLong::NO_ARGUMENT]
@@ -204,9 +203,11 @@ class AppConfig
           @yes = false
         when '--force-pin'
           @force_pin = true
+        when '--force-no-pin'
+          @force_pin = false
         end
       }
-    rescue getoptlong_ambiguousoption, GetoptLong::NeedlessArgument,
+    rescue GetoptLong::AmbiguousOption, GetoptLong::NeedlessArgument,
         GetoptLong::MissingArgument, GetoptLong::InvalidOption
       usage
       exit 1
@@ -215,7 +216,9 @@ class AppConfig
     if ! $stdout.isatty
       @quiet = true
       @yes = false if @yes.nil?
-      @force_pin = true
+      @force_pin = true if @force_pin.nil?
+    else
+      @force_pin = false if @force_pin.nil?
     end
 
     # warn the user about unknown (possibly misspelled) severities
@@ -426,6 +429,10 @@ class Viewer
             a = "n" if ! @config.yes
           end
         end
+        if a.nil?
+          a = "got nil!"
+          @config.frontend.puts ""
+        end
         if a == ""
           if hold_pkg_keys.empty?
             answer = "y"
@@ -916,9 +923,10 @@ module Factory
       # TRANSLATORS: this sentence, followed by the translation of "Done" (see above) should fit in less than 79 columns to work well with default width terminals
       yield _("Retrieving bug reports..."), "0%"
       begin
+        config.frontend.puts " Using QueryStep = #{config.querystep}" if $DEBUG
         config.frontend.puts " Using ParseStep = #{config.parsestep}" if $DEBUG
         # obtain the array of bugs
-        bugs = config.parser.parse(ma_copies, config.parsestep, config.severity) { |pct|
+        bugs = config.parser.parse(ma_copies, config.querystep, config.parsestep, config.severity) { |pct|
           yield _("Retrieving bug reports..."), pct
         }
       rescue SOAP::HTTPStreamError => exception
@@ -927,6 +935,7 @@ module Factory
         $stderr.puts _("E: ") + _("HTTP GET failed")
         retrycount -= 1
         if config.frontend.yes_or_no?(_("Retry downloading bug information?")) && retrycount > 0
+          config.querystep = 1 if config.querystep != 1 && config.frontend.yes_or_no?(_("One package at a time?"))
           config.parsestep = 1 if config.parsestep != 1 && config.frontend.yes_or_no?(_("One bug report at a time?"))
           retry
         end
@@ -938,6 +947,7 @@ module Factory
         $stderr.puts _("E: ") + _("Empty stream from SOAP")
         retrycount -= 1
         if config.frontend.yes_or_no?(_("Retry downloading bug information?")) && retrycount > 0
+          config.querystep = 1 if config.querystep != 1 && config.frontend.yes_or_no?(_("One package at a time?"))
           config.parsestep = 1 if config.parsestep != 1 && config.frontend.yes_or_no?(_("One bug report at a time?"))
           retry
         end
@@ -955,6 +965,7 @@ module Factory
         end
         retrycount -= 1
         if config.frontend.yes_or_no?(_("Retry downloading bug information?")) && retrycount > 0
+          config.querystep = 1 if config.querystep != 1 && config.frontend.yes_or_no?(_("One package at a time?"))
           config.parsestep = 1 if config.parsestep != 1 && config.frontend.yes_or_no?(_("One bug report at a time?"))
           retry
         end
@@ -1149,6 +1160,15 @@ module Factory
         new_ver = new_pkgs[pkg_key]["version"] if new_pkgs[pkg_key] != nil
         cur_ver = cur_pkgs[pkg_key]["version"] if cur_pkgs[pkg_key] != nil
 
+        # treat back-ported versions as if they were the corresponding
+        # official versions (this special handling is necessary until
+        # we can query the BTS version tracking directly, see #694979#19,
+        # and until back-ports are correctly handled by BTS version tracking)
+        new_o_ver = nil
+        cur_o_ver = nil
+        new_o_ver = new_ver.gsub(/~bpo[0-9].*$/, '') if new_ver != nil
+        cur_o_ver = cur_ver.gsub(/~bpo[0-9].*$/, '') if cur_ver != nil
+
         # show progress
         yield _("Parsing Found/Fixed information..."),
         "#{(i.to_f/max.to_f*100).to_i}%" if (i % step) == 0
@@ -1159,7 +1179,7 @@ module Factory
         src_name = new_pkgs[pkg_key]
         src_name = src_name["package"] if src_name != nil
 
-        val = true if bug_is_irrelevant(src_name, cur_ver, new_ver,
+        val = true if bug_is_irrelevant(src_name, cur_o_ver, new_o_ver,
                                         bug.bug_number, bug.fixed, bug.found, bug.stat)
         val
       }
@@ -1246,6 +1266,8 @@ class ConsoleFrontend
         return true
       elsif a == "N" || a == "n"
         return false
+      elsif a.nil?
+        $stdout.print "\n"
       end
     end
   end
